@@ -2,7 +2,7 @@
 ** PSInterpreter.cpp                                                    **
 **                                                                      **
 ** This file is part of dvisvgm -- a fast DVI to SVG converter          **
-** Copyright (C) 2005-2019 Martin Gieseking <martin.gieseking@uos.de>   **
+** Copyright (C) 2005-2020 Martin Gieseking <martin.gieseking@uos.de>   **
 **                                                                      **
 ** This program is free software; you can redistribute it and/or        **
 ** modify it under the terms of the GNU General Public License as       **
@@ -24,11 +24,13 @@
 #include <sstream>
 #include <unordered_map>
 #include "FileFinder.hpp"
+#include "FileSystem.hpp"
 #include "InputReader.hpp"
 #include "Message.hpp"
 #include "PSFilter.hpp"
 #include "PSInterpreter.hpp"
 #include "SignalHandler.hpp"
+#include "utility.hpp"
 
 using namespace std;
 
@@ -254,47 +256,50 @@ void PSInterpreter::callActions (InputReader &in) {
 		void (PSActions::*handler)(vector<double> &p);  // operation handler
 	};
 	static const unordered_map<string, Operator> operators {
-		{"applyscalevals", { 3, &PSActions::applyscalevals}},
-		{"clip",           { 0, &PSActions::clip}},
-		{"clippath",       { 0, &PSActions::clippath}},
-		{"closepath",      { 0, &PSActions::closepath}},
-		{"curveto",        { 6, &PSActions::curveto}},
-		{"eoclip",         { 0, &PSActions::eoclip}},
-		{"eofill",         { 0, &PSActions::eofill}},
-		{"fill",           { 0, &PSActions::fill}},
-		{"grestore",       { 0, &PSActions::grestore}},
-		{"grestoreall",    { 0, &PSActions::grestoreall}},
-		{"gsave",          { 0, &PSActions::gsave}},
-		{"initclip",       { 0, &PSActions::initclip}},
-		{"lineto",         { 2, &PSActions::lineto}},
-		{"makepattern",    {-1, &PSActions::makepattern}},
-		{"moveto",         { 2, &PSActions::moveto}},
-		{"newpath",        { 1, &PSActions::newpath}},
-		{"querypos",       { 2, &PSActions::querypos}},
-		{"raw",            {-1, nullptr}},
-		{"restore",        { 1, &PSActions::restore}},
-		{"rotate",         { 1, &PSActions::rotate}},
-		{"save",           { 1, &PSActions::save}},
-		{"scale",          { 2, &PSActions::scale}},
-		{"setblendmode",   { 1, &PSActions::setblendmode}},
-		{"setcmykcolor",   { 4, &PSActions::setcmykcolor}},
-		{"setdash",        {-1, &PSActions::setdash}},
-		{"setgray",        { 1, &PSActions::setgray}},
-		{"sethsbcolor",    { 3, &PSActions::sethsbcolor}},
-		{"setlinecap",     { 1, &PSActions::setlinecap}},
-		{"setlinejoin",    { 1, &PSActions::setlinejoin}},
-		{"setlinewidth",   { 1, &PSActions::setlinewidth}},
-		{"setmatrix",      { 6, &PSActions::setmatrix}},
-		{"setmiterlimit",  { 1, &PSActions::setmiterlimit}},
-		{"setnulldevice",  { 1, &PSActions::setnulldevice}},
-		{"setopacityalpha",{ 1, &PSActions::setopacityalpha}},
-		{"setshapealpha",  { 1, &PSActions::setshapealpha}},
-		{"setpagedevice",  { 0, &PSActions::setpagedevice}},
-		{"setpattern",     {-1, &PSActions::setpattern}},
-		{"setrgbcolor",    { 3, &PSActions::setrgbcolor}},
-		{"shfill",         {-1, &PSActions::shfill}},
-		{"stroke",         { 0, &PSActions::stroke}},
-		{"translate",      { 2, &PSActions::translate}},
+		{"applyscalevals",         { 3, &PSActions::applyscalevals}},
+		{"clip",                   { 0, &PSActions::clip}},
+		{"clippath",               { 0, &PSActions::clippath}},
+		{"closepath",              { 0, &PSActions::closepath}},
+		{"curveto",                { 6, &PSActions::curveto}},
+		{"eoclip",                 { 0, &PSActions::eoclip}},
+		{"eofill",                 { 0, &PSActions::eofill}},
+		{"fill",                   { 0, &PSActions::fill}},
+		{"grestore",               { 0, &PSActions::grestore}},
+		{"grestoreall",            { 0, &PSActions::grestoreall}},
+		{"gsave",                  { 0, &PSActions::gsave}},
+		{"image",                  { 3, &PSActions::image}},
+		{"initclip",               { 0, &PSActions::initclip}},
+		{"lineto",                 { 2, &PSActions::lineto}},
+		{"makepattern",            {-1, &PSActions::makepattern}},
+		{"moveto",                 { 2, &PSActions::moveto}},
+		{"newpath",                { 1, &PSActions::newpath}},
+		{"querypos",               { 2, &PSActions::querypos}},
+		{"raw",                    {-1, nullptr}},
+		{"restore",                { 1, &PSActions::restore}},
+		{"rotate",                 { 1, &PSActions::rotate}},
+		{"save",                   { 1, &PSActions::save}},
+		{"scale",                  { 2, &PSActions::scale}},
+		{"setblendmode",           { 1, &PSActions::setblendmode}},
+		{"setcolorspace",          { 1, &PSActions::setcolorspace}},
+		{"setcmykcolor",           { 4, &PSActions::setcmykcolor}},
+		{"setdash",                {-1, &PSActions::setdash}},
+		{"setfillconstantalpha",   { 1, &PSActions::setfillconstantalpha}},
+		{"setgray",                { 1, &PSActions::setgray}},
+		{"sethsbcolor",            { 3, &PSActions::sethsbcolor}},
+		{"setisshapealpha",        { 1, &PSActions::setisshapealpha}},
+		{"setlinecap",             { 1, &PSActions::setlinecap}},
+		{"setlinejoin",            { 1, &PSActions::setlinejoin}},
+		{"setlinewidth",           { 1, &PSActions::setlinewidth}},
+		{"setmatrix",              { 6, &PSActions::setmatrix}},
+		{"setmiterlimit",          { 1, &PSActions::setmiterlimit}},
+		{"setnulldevice",          { 1, &PSActions::setnulldevice}},
+		{"setpagedevice",          { 0, &PSActions::setpagedevice}},
+		{"setpattern",             {-1, &PSActions::setpattern}},
+		{"setrgbcolor",            { 3, &PSActions::setrgbcolor}},
+		{"setstrokeconstantalpha", { 1, &PSActions::setstrokeconstantalpha}},
+		{"shfill",                 {-1, &PSActions::shfill}},
+		{"stroke",                 { 0, &PSActions::stroke}},
+		{"translate",              { 2, &PSActions::translate}},
 	};
 	if (_actions) {
 		in.skipSpace();
@@ -352,7 +357,7 @@ int GSDLLCALL PSInterpreter::error (void *inst, const char *buf, int len) {
 /** Returns the total number of pages of a PDF file.
  *  @param[in] fname name/path of the PDF file */
 int PSInterpreter::pdfPageCount (const string &fname) {
-	executeRaw("\n("+fname+")@pdfpagecount ", 1);
+	executeRaw("\n("+FileSystem::ensureForwardSlashes(fname)+")@pdfpagecount ", 1);
 	if (!_rawData.empty()) {
 		size_t index;
 		int ret = stoi(_rawData[0], &index, 10);
@@ -370,10 +375,84 @@ int PSInterpreter::pdfPageCount (const string &fname) {
  *  @return the bounding box of the given page */
 BoundingBox PSInterpreter::pdfPageBox (const string &fname, int pageno) {
 	BoundingBox pagebox;
-	executeRaw("\n"+to_string(pageno)+"("+fname+")@pdfpagebox ", 4);
+	executeRaw("\n"+to_string(pageno)+"("+FileSystem::ensureForwardSlashes(fname)+")@pdfpagebox ", 4);
 	if (_rawData.size() < 4)
 		pagebox.invalidate();
 	else
 		pagebox = BoundingBox(stod(_rawData[0]), stod(_rawData[1]), stod(_rawData[2]), stod(_rawData[3]));
 	return pagebox;
 }
+
+
+vector<PSDeviceInfo> PSInterpreter::getImageDeviceInfos () {
+	vector<PSDeviceInfo> infos {
+		{"none", "no processing of bitmap images"},
+		{"jpeg", "color JPEG format"},
+		{"jpeggray", "grayscale JPEG format"},
+		{"png", "grayscale or 24-bit color PNG format"},
+		{"pnggray", "grayscale PNG format"},
+		{"pngmono", "black-and-white PNG format"},
+		{"pngmonod", "dithered black-and-white PNG format"},
+		{"png16", "4-bit color PNG format"},
+		{"png256", "8-bit color PNG format"},
+		{"png16m", "24-bit color PNG format"},
+	};
+	return infos;
+}
+
+
+void PSInterpreter::listImageDeviceInfos (ostream &os) {
+	for (const PSDeviceInfo &info : getImageDeviceInfos())
+		os << setw(8) << left << info.name << " | " << info.description << '\n';
+}
+
+
+/** Returns true if a given PS device name is known. The function deosn't
+ *  check whether the device is actually available.
+ *  @param[in] deviceStr device specifier of the form <device name>[:<param>] */
+bool PSInterpreter::imageDeviceKnown (string deviceStr) {
+	if (deviceStr.empty() || !isalpha(deviceStr[0]))
+		return false;
+	deviceStr = deviceStr.substr(0, deviceStr.find(':'));  // strip optional argument
+	auto infos = getImageDeviceInfos();
+	auto it = find_if(infos.begin(), infos.end(), [&](PSDeviceInfo &info) {
+		return info.name == deviceStr;
+	});
+	return it != infos.end();
+}
+
+
+/** Sets the output device used to create bitmap images.
+ *  @param[in] deviceStr device specifier of the form <device name>[:<param>]
+ *  @return true on success, false if device is not supported */
+bool PSInterpreter::setImageDevice (const string &deviceStr) {
+	auto params = util::split(deviceStr, ":");
+	string name = util::tolower(params[0]);
+	if (!imageDeviceKnown(name))
+		return false;
+	if (name != "jpeg" && name != "png" && name != "none") {
+		// check if image device is supported by Ghostscript
+		executeRaw("devicedict/"+name+" known{1}{0}ifelse\n", 1);
+		if (_rawData.empty() || _rawData[0] != "1")
+			throw PSException("output device '"+name+"' is not available");
+	}
+	string ps = "/@imgdevice("+name+")store ";
+	try {
+		if (params.size() > 1) {
+			// set JPEG quality level if given
+			if (name.substr(0, 4) == "jpeg") {
+				int quality = max(0, min(stoi(params[1]), 100));
+				ps += "/JPEGQ "+to_string(quality)+" def ";
+			}
+			else if (name == "pngmonod") {
+				int minFeatureSize = max(0, min(stoi(params[1]), 4));
+				ps += "/MinFeatureSize "+to_string(minFeatureSize)+" def ";
+			}
+		}
+	}
+	catch (...) {
+		throw PSException("invalid device option '"+params[1]+"' (integer expected)");
+	}
+	execute(ps);
+	return true;
+}
