@@ -2,7 +2,7 @@
 ** Subfont.cpp                                                          **
 **                                                                      **
 ** This file is part of dvisvgm -- a fast DVI to SVG converter          **
-** Copyright (C) 2005-2019 Martin Gieseking <martin.gieseking@uos.de>   **
+** Copyright (C) 2005-2020 Martin Gieseking <martin.gieseking@uos.de>   **
 **                                                                      **
 ** This program is free software; you can redistribute it and/or        **
 ** modify it under the terms of the GNU General Public License as       **
@@ -97,10 +97,11 @@ Subfont* SubfontDefinition::subfont (const string &id) const {
 
 
 /** Returns all subfonts defined in this SFD. */
-int SubfontDefinition::subfonts (vector<Subfont*> &sfs) const {
+vector<Subfont*> SubfontDefinition::subfonts () const {
+	vector<Subfont*> subfonts;
 	for (const auto &strsfpair : _subfonts)
-		sfs.push_back(strsfpair.second.get());
-	return int(sfs.size());
+		subfonts.push_back(strsfpair.second.get());
+	return subfonts;
 }
 
 //////////////////////////////////////////////////////////////////////
